
------------------------------------------------------------------------
--
-- AbsorbsViewer
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


local AbsorbsMonitor = _G.AbsorbsMonitor;

AbsorbsViewer = LibStub("AceAddon-3.0"):NewAddon("AbsorbsViewer", "AceConsole-3.0", "AceEvent-3.0");

local AbsorbEffects;
local EffectNames;

local baseFrame;
local framePool_Units = {};
local framePool_Effects = {};
local activeUnitsMap = {};
local activeUnitsList = {};
local lowestUnitFrame = nil;

local isHidden;


-- #################################################################################################


local function GetUnitClassColorRGB(name)
	local _, class = UnitClass(name);

	if _G.RAID_CLASS_COLORS and _G.RAID_CLASS_COLORS[class] then
		return _G.RAID_CLASS_COLORS[class].r, _G.RAID_CLASS_COLORS[class].g, _G.RAID_CLASS_COLORS[class].b;
	else
		return 0.63, 0.63, 0.63;	
	end
end

local function OnDragStart(self)
	if IsAltKeyDown() then
		self.isMoving = true;
		self:StartMoving();
	end
end

local function OnDragStop(self)
	if self.isMoving then
		self.isMoving = false;
		self:StopMovingOrSizing();

		local x, y = self:GetLeft(), self:GetTop();
	end
end

local function OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
	GameTooltip:SetText("ALT + Drag to move the frame anchor", nil, nil, nil, nil, 1);
end

local function OnLeave(self)
	GameTooltip:Hide();	
end

local backdrop = {bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeSize = 0.90,
		insets = {left = 1, right = 1, top = 1, bottom = 1}};
		
-- Note this function is exactly INVERSE to SortEffects in Core.lua
local function SortEffects(a, b)
	if a.priority == b.priority then
		return (a.amount > b.amount);
	else
		return (a.priority < b.priority);
	end
end

local function RepositionUnits()
	for i, frame in pairs(activeUnitsList) do
		frame:ClearAllPoints();
		
		if(i > 1) then
			frame:SetPoint("TOPLEFT", activeUnitsList[i - 1], "BOTTOMLEFT", 0, -2);
		else
			frame:SetPoint("TOPLEFT", baseFrame, "BOTTOMLEFT", 0, -2);
		end
	end
	
	lowestUnitFrame = activeUnitsList[#activeUnitsList];
end

local function RepositionEffects(guid)
	if(not activeUnitsMap[guid]) then return; end
	
	local effectsList = activeUnitsMap[guid].effectsList;
	
	table.sort(effectsList, SortEffects);
	
	for i, frame in pairs(effectsList) do
		frame:ClearAllPoints();
	
		if(i > 1) then
			frame:SetPoint("TOPLEFT", effectsList[i - 1], "BOTTOMLEFT", 0, -2);
		else
			frame:SetPoint("TOPLEFT", activeUnitsMap[guid], "TOPLEFT", 10, -22 );
		end
	end
end

local function CreateBaseFrame()
	baseFrame = CreateFrame("Frame", "absorbsview_baseFrame", UIParent);
	
	baseFrame:SetHeight(20);
	baseFrame:SetMovable(true);
	baseFrame:EnableMouse(true);
	baseFrame:SetClampedToScreen(true);
	baseFrame:RegisterForDrag("LeftButton");
	baseFrame:SetBackdrop(backdrop);
	baseFrame:SetBackdropColor(0, 0, 0, 0.5);
	baseFrame:SetBackdropBorderColor(1, 1, 1, 0.6);
	baseFrame:SetScript("OnDragStart", OnDragStart);
	baseFrame:SetScript("OnDragStop", OnDragStop);
	baseFrame:SetScript("OnShow", OnShow);
	baseFrame:SetScript("OnEnter", OnEnter);
	baseFrame:SetScript("OnLeave", OnLeave);
	baseFrame.name = "absorbs_viewer";
	
	baseFrame.text = baseFrame:CreateFontString(nil, "OVERLAY");
	baseFrame.text:SetPoint("LEFT", baseFrame, 10, 0);
	baseFrame.text:SetFontObject(GameFontHighlightSmall);
	baseFrame.text:SetText("AbsorbsMonitor Effect Viewer");
	
	baseFrame:SetScale(1.0);
	
	baseFrame:SetWidth(225);	
	
	baseFrame:SetAlpha(1.0);
	
	baseFrame:ClearAllPoints();
	baseFrame:SetPoint("CENTER", UIParent, "CENTER");
	
	baseFrame.CloseButton = CreateFrame("Button", nil, baseFrame);
	baseFrame.CloseButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up.blp");
	baseFrame.CloseButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down.blp");
	baseFrame.CloseButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");
	baseFrame.CloseButton:SetWidth(20);
	baseFrame.CloseButton:SetHeight(20);
	baseFrame.CloseButton:SetPoint("TOPRIGHT", baseFrame, "TOPRIGHT", 0, 0);
	baseFrame.CloseButton:SetScript("OnClick", function(self) self:GetParent():Hide(); isHidden = true; end);
	
	lowestUnitFrame = baseFrame;
end

local function GetUnitFrame()
	if(#framePool_Units > 0) then
		return table.remove(framePool_Units, 1);
	end
	
	local frame = CreateFrame("Frame", nil, baseFrame);
	
	frame:SetBackdrop(backdrop);
	frame:SetBackdropColor(0, 0, 0, 0.5);
	frame:SetBackdropBorderColor(1, 1, 1, 0.3);	
	
	local path, size, style = GameFontHighlight:GetFont();	
	
	frame.textName = frame:CreateFontString(nil, "OVERLAY");
	frame.textName:SetPoint("TOPLEFT", frame, 10, -5);
	frame.textName:SetFontObject(GameFontHighlightSmall);
	frame.textName:SetFont(path, size, style);
	
	frame.textTotal = frame:CreateFontString(nil, "OVERLAY");
	frame.textTotal:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -10, -5);
	frame.textTotal:SetFontObject(GameFontHighlightSmall);
	frame.textTotal:SetFont(path, size, style);
	
	frame:SetScale(1.0);
	
	frame:SetWidth(225);	
	
	frame:SetAlpha(1.0);
	
	frame:ClearAllPoints();
	
	return frame;
end

local function RemoveUnit(guid)
	if(not activeUnitsMap[guid]) then return; end
	
	local unitFrame = activeUnitsMap[guid];
	
	activeUnitsMap[guid] = nil;
	
	for i, frame in pairs(activeUnitsList) do
		if(frame == unitFrame) then
			table.remove(activeUnitsList, i);
			break;
		end
	end
		
	unitFrame.effectsMap = nil;
	unitFrame.effectsList = nil;
	
	unitFrame:Hide();
	table.insert(framePool_Units, unitFrame);
	
	
	if(#activeUnitsList > 0) then
		RepositionUnits();
	else
		lowestUnitFrame = baseFrame;
	end
end

local function GetEffectFrame()
	if(#framePool_Effects > 0) then
		return table.remove(framePool_Effects, 1);
	end
		
	local frame = CreateFrame("StatusBar", nil, baseFrame);
	frame:Hide();		
	
	frame:SetClampedToScreen(true);
	frame:SetMinMaxValues(0, 1);
		
	frame.bg = CreateFrame("StatusBar", nil, frame);
	frame.bg:SetMinMaxValues(0, 1);
	frame.bg:SetValue(1.0);
	frame.bg:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0);
	frame.bg:SetFrameLevel(0);
	frame.bg:SetStatusBarTexture([[Interface\TargetingFrame\UI-StatusBar]]);
	frame.bg:SetWidth(205);
	frame.bg:SetHeight(14);
	frame.bg:SetStatusBarColor(0.5, 0.5, 0.5, 105);
	
	frame:SetStatusBarTexture([[Interface\TargetingFrame\UI-StatusBar]]);	
	frame:SetWidth(205);
	frame:SetHeight(14);
	frame:SetScale(1.0);	
	
	local path, size, style = GameFontHighlight:GetFont();	
	size = size - 2;
	
	frame.textName = frame:CreateFontString(nil, "OVERLAY");
	frame.textName:SetFontObject(GameFontHighlight);
	frame.textName:SetJustifyH("LEFT");
	frame.textName:SetPoint("LEFT", frame, "LEFT", 0, 0);
	frame.textName:SetFont(path, size, style);	
	frame.textName:SetHeight(14);
	frame.textName:SetWidth(125);	
			
	frame.textAmount = frame:CreateFontString(nil, "OVERLAY");
	frame.textAmount:SetFontObject(GameFontHighlight);
	frame.textAmount:SetJustifyH("RIGHT");
	frame.textAmount:SetPoint("RIGHT", frame, "RIGHT", -3, 0);
	frame.textAmount:SetFont(path, size, style);
	frame.textAmount:SetHeight(10);
	frame.textAmount:SetWidth(80);
		
	frame:SetAlpha(1.0);
	
	return frame;
end

local function ReleaseEffectFrame(frame)
	frame:Hide();
	
	table.insert(framePool_Effects, frame);
end

local function AddEffect(guid, unitName, spellId, amount, quality)
	local unitFrame, effectFrame;

	if(not activeUnitsMap[guid]) then
		unitFrame = GetUnitFrame();
		
		unitFrame:SetHeight(30);
		
		unitFrame.textName:SetText(unitName);
		unitFrame.textName:SetTextColor(GetUnitClassColorRGB(unitName));
		
		activeUnitsMap[guid] = unitFrame;
		table.insert(activeUnitsList, unitFrame);
		
		unitFrame:SetPoint("TOPLEFT", lowestUnitFrame, "BOTTOMLEFT", 0, -2);
		lowestUnitFrame = unitFrame;
		
		unitFrame:Show();
	else
		unitFrame = activeUnitsMap[guid];
	end
	
	effectFrame = GetEffectFrame();
	
	effectFrame.spellId = spellId;
	
	effectFrame:SetValue(1.0);
	effectFrame.amount = amount;
	effectFrame.max = amount;
	effectFrame.priority = AbsorbEffects[spellId][1];
	
	effectFrame:SetStatusBarColor(1 - quality, quality, 0, 1.0);
	
	if(EffectNames[spellId]) then
		effectFrame.textName:SetText(EffectNames[spellId]);
	else
		effectFrame.textName:SetText("SpellId "..spellId);
	end
	
	if(amount > 0) then
		effectFrame.textAmount:SetText(amount);
	end
	
	if(not unitFrame.effectsMap) then
		effectFrame:SetPoint("TOPLEFT", unitFrame, "TOPLEFT", 10, -22);
				
		unitFrame.effectsMap = {[spellId] = effectFrame};
		unitFrame.effectsList = {effectFrame};
	else		
		unitFrame.effectsMap[spellId] = effectFrame;			
		table.insert(unitFrame.effectsList, effectFrame);
		
		RepositionEffects(guid);
	end		
		
	unitFrame:SetHeight(unitFrame:GetHeight() + 14);
	effectFrame:Show();
end

local function RemoveEffect(guid, spellId)
	if(activeUnitsMap[guid] and activeUnitsMap[guid].effectsMap[spellId]) then
		local unitFrame = activeUnitsMap[guid];
	
		ReleaseEffectFrame(unitFrame.effectsMap[spellId]);
		
		unitFrame.effectsMap[spellId] = nil;
		
		for i, frame in pairs(unitFrame.effectsList) do
			if(frame.spellId == spellId) then
				table.remove(unitFrame.effectsList, i);
				break;
			end
		end
		
		if(#(unitFrame.effectsList) == 0) then			
			RemoveUnit(guid);
		else
			RepositionEffects(guid);
			unitFrame:SetHeight(unitFrame:GetHeight() - 14);
		end
	end		
end

local function UpdateEffect(guid, spellId, amount)
	if(amount == 0) then
		RemoveEffect(guid, spellId);
	end

	if(activeUnitsMap[guid] and activeUnitsMap[guid].effectsMap[spellId]) then
		local effectFrame = activeUnitsMap[guid].effectsMap[spellId];
		
		effectFrame.amount = amount;
		
		if(amount >= effectFrame.max) then
			effectFrame:SetValue(1.0);
			effectFrame.max = amount;
			
			-- We might need to reposition now
			RepositionEffects(guid);
		else
			effectFrame:SetValue(amount / effectFrame.max);			
		end
		
		effectFrame.textAmount:SetText(amount);
	end
end


-- #################################################################################################


function AbsorbsViewer:Absorbs_EffectApplied(obj, sourceGUID, sourceName, destGUID, destName, spellId, amount, quality)
	AddEffect(destGUID, destName, spellId, amount, quality);
end

function AbsorbsViewer:Absorbs_EffectUpdated(obj, guid, spellId, amount)
	UpdateEffect(guid, spellId, amount);
end

function AbsorbsViewer:Absorbs_EffectRemoved(obj, guid, spellId)
	RemoveEffect(guid, spellId);
end

function AbsorbsViewer:Absorbs_UnitUpdated(obj, guid, amount, quality)
	if(activeUnitsMap[guid]) then
		local textTotal = activeUnitsMap[guid].textTotal;
		
		textTotal:SetText(amount);
		textTotal:SetTextColor(1 - quality, quality, 0);
	end
end

function AbsorbsViewer:Absorbs_UnitCleared(obj, guid)
	if(activeUnitsMap[guid]) then
		-- no need to worry about any data structures here
		for i, frame in pairs(activeUnitsMap[guid].effectsList) do
			ReleaseEffectFrame(frame);
		end
		
		RemoveUnit(guid);
	end
end

local areaSpellMapping = {};

-- I know you can have only one area effect per unit active in this implementation,
-- but more than one is impossible ingame anyway
function AbsorbsViewer:Absorbs_AreaCreated(obj, sourceGUID, sourceName, triggerGUID, spellId, value, quality)
	AddEffect(triggerGUID, "Area Effects by "..sourceName, spellId, value, quality);
	
	areaSpellMapping[triggerGUID] = spellId;
end

-- triggerGUID, value
function AbsorbsViewer:Absorbs_AreaUpdated(obj, triggerGUID, value)
	UpdateEffect(triggerGUID, areaSpellMapping[triggerGUID], value);
end

function AbsorbsViewer:Absorbs_AreaCleared(obj, triggerGUID)
	RemoveEffect(triggerGUID, areaSpellMapping[triggerGUID]);
	
	areaSpellMapping[triggerGUID] = nil;
end

function AbsorbsViewer:OnSlashCommand()
	baseFrame:Show();
	isHidden = false;
end

function AbsorbsViewer:OnEnable()
	CreateBaseFrame();
	baseFrame:Hide();
	isHidden = true;
	
	AbsorbEffects = AbsorbsMonitor.Effects;
	EffectNames = AbsorbsMonitor.EffectNames;
	
	self:RegisterMessage("Absorbs_EffectApplied");
	self:RegisterMessage("Absorbs_EffectUpdated");
	self:RegisterMessage("Absorbs_EffectRemoved");
	
	self:RegisterMessage("Absorbs_AreaCreated");
	self:RegisterMessage("Absorbs_AreaUpdated");
	self:RegisterMessage("Absorbs_AreaCleared");
	
	self:RegisterMessage("Absorbs_UnitUpdated");
	self:RegisterMessage("Absorbs_UnitCleared");
	
	self:RegisterChatCommand("av", "OnSlashCommand");
end

function AbsorbsViewer:OnDisable()
end
